import argparse
from . import config, utils
from .cli import run_cli
from .smoke import run_smoke


def main():
    parser = argparse.ArgumentParser(
        prog="whis",
        description="Suggest a shell command from a natural-language prompt.",
        epilog=(
            "Usage:\n"
            "  1. Write your request in natural language\n"
            "  2. After a suggestion appears, choose:\n"
            "     • [enter] quit+paste the command into your terminal\n"
            "     • [r] retry\n"
            "     • refine the command by writing a feedback message\n"
            "\n"
            "Note: After [enter], the command is automatically pasted into your terminal."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("--version", action="store_true", help="Show version.")
    parser.add_argument("--show-config", action="store_true", help="Show configuration details.")

    parser.add_argument(  # $ whis -o "list mp3 files"
        "-o",
        "--oneshot",
        nargs="+",
        metavar="WORD",
        help="Run one-shot mode (no interactive session).",
    )

    subparsers = parser.add_subparsers(dest="command")

    subparsers.add_parser("run", help="Start interactive session")
    # subparsers.add_parser("setup", help="Run interactive configuration")
    subparsers.add_parser("smoke", help="Run a quick smoke test")

    args = parser.parse_args()

    if args.version:
        print(config.get_version())
        return

    if args.show_config:
        utils.print_config()
        return

    if args.oneshot:
        prompt = " ".join(args.oneshot)
        run_cli(oneshot=prompt)
        return

    cmd = args.command or "run"
    if cmd == "run":
        run_cli()
    elif cmd == "smoke":
        run_smoke()


if __name__ == "__main__":
    if config.is_dev:
        utils.print_config()
    main()



import logging
from abc import ABC, abstractmethod
from enum import Enum

from . import config

try:
    from openai import OpenAI

    OPENAI_AVAILABLE = True
except ImportError:
    OPENAI_AVAILABLE = False

try:
    import ollama

    OLLAMA_AVAILABLE = True
except ImportError:
    OLLAMA_AVAILABLE = False

logger = logging.getLogger(__name__)


class ProviderType(Enum):
    OLLAMA = "ollama"
    OPENAI = "openai"
    # ANTHROPIC = "anthropic"  # todo
    # GEMINI = "gemini"  # todo


class ProviderRegistry:
    def __init__(self):
        self._providers = {}

    class UnavailableProviderError(Exception):
        pass

    def register(self, cls):
        logger.debug("Registering provider: %s", cls)
        self._providers[cls.name] = cls
        return cls

    def _get_provider_class(self, provider_name):
        try:
            return self._providers[provider_name]
        except KeyError:
            raise self.UnavailableProviderError(f"Provider {provider_name} is not available.")

    def create(self, provider_name, model):
        # todo manage errors
        return self._get_provider_class(provider_name)(model)

    def create_by_env(self):
        return self.create(config.llm_provider, config.llm_model)

    def list(self) -> list[str]:
        return sorted(self._providers.keys())


registry = ProviderRegistry()


class LLMProvider(ABC):
    name = None
    label = None
    temp = None

    def __init__(self, model):
        self.history = []
        self.model = model
        self._set_system_prompt()

    def _set_system_prompt(self) -> None:
        self._add_message("system", config.SYSTEM_PROMPT)

    def _add_message(self, role: str, content: str):
        logger.debug("Adding message to history: role=%s, content=%s", role, content)
        self.history.append({"role": role, "content": content})

    def _log_recent_history(self, max_messages: int = 5) -> None:
        if len(self.history) <= 1:
            return

        user_messages = self.history[1:]  # skip system prompt
        recent = user_messages[-max_messages:] if len(user_messages) >= max_messages else user_messages

        logger.debug("Last messages in history:")
        for msg in recent:
            logger.debug("  Role=%s, Content=%s", msg["role"], msg["content"])

    def submit(self) -> str:
        logger.debug("Submitting request to LLM")
        self._log_recent_history()
        response = self._submit()
        self._add_message("assistant", response)
        logger.debug("Response: %s", response)
        return response

    @abstractmethod
    def _submit(self):
        pass

    def say(self, message) -> str:
        self._add_message("user", message)
        response = self.submit()
        command = self.parse_command(response)
        return command

    @abstractmethod
    def is_available(self) -> bool:
        pass

    @staticmethod
    def parse_command(command: str) -> str:
        if command.startswith("```") and command.endswith("```"):
            lines = command.splitlines()
            command_lines = lines[1:-1]
            command = "\n".join(command_lines)
        else:
            command = command.strip("`")
        return command.strip()

    def __str__(self):
        return f"{self.label} ({self.model})"


@registry.register
class OpenAIProvider(LLMProvider):
    name = "openai"
    label = "OpenAI"
    temp = 0.2

    def __init__(self, model: str):
        if not OPENAI_AVAILABLE:
            raise ImportError("Missing package `openai`.")

        super().__init__(model)

        self.model = model
        self.provider = OpenAI()

    def is_available(self) -> bool:
        return OPENAI_AVAILABLE

    def _submit(self) -> str:
        response = self.provider.chat.completions.create(
            model=self.model,
            messages=self.history,
            temperature=self.temp,
        )
        return response.choices[0].message.content.strip()


@registry.register
class OllamaProvider(LLMProvider):
    name = "ollama"
    label = "Ollama"
    temp = 0.2

    def __init__(self, model: str):
        if not OLLAMA_AVAILABLE:
            raise ImportError("Missing package `ollama`.")
        super().__init__(model)

        self.model = model
        self.provider = ollama.Client()

    def is_available(self) -> bool:
        return OLLAMA_AVAILABLE

    def _submit(self) -> str:
        response = self.provider.chat(
            model=self.model,
            messages=self.history,
            options={"temperature": self.temp},
        )
        return response.message.content.strip()


class DummyProvider(LLMProvider):
    """Dummy provider for testing etc..."""

    name = "dummy"
    label = "Dummy"
    temp = 0.0

    def __init__(self, model: str = "dummy"):
        super().__init__(model)

    def is_available(self) -> bool:
        return True

    def _submit(self) -> str:
        return "a dummy response"


if config.is_dev or config.is_test or config.llm_provider == "dummy":
    registry.register(DummyProvider)



from .providers import registry
from .utils import ANSIColors, colorize


def run_smoke():
    print("Running smoke test...")
    provider = registry.create_by_env()
    if provider.is_available():
        print("Provider is available.", end=" ")
        print(provider)
    print("Asking for suggestion...", end=" ")
    print(colorize("list mp3 files", ANSIColors.BOLD_ORANGE))
    sug = provider.say("list mp3 files")
    print(f"Suggestion received: {colorize(sug, ANSIColors.BOLD_CYAN)}")
    print("OK. No errors found.")
    return



import logging
import os

try:
    import tomllib
except ImportError:  # py <= 3.10
    import tomli as tomllib
from datetime import datetime
from enum import Enum
from importlib import metadata
from pathlib import Path

try:
    from dotenv import load_dotenv
except ImportError:
    pass  # ok if dotenv is not installed
else:
    load_dotenv()


class Env(str, Enum):
    DEV = "dev"
    PROD = "prod"
    TEST = "test"


ENV = os.environ.get("WHIS_ENV", Env.PROD).lower()

is_dev = ENV == Env.DEV
is_prod = ENV == Env.PROD
is_test = ENV == Env.TEST

REPO_ROOT = Path(__file__).resolve().parents[1]
PACKAGE_ROOT = REPO_ROOT / "whis"
SYSTEM_PROMPT_FILE = PACKAGE_ROOT / "system_prompt.txt"
CONFIG_FILE = None

if is_dev:
    LOG_FILE = REPO_ROOT / "whis.dev.log"
    config_toml = {}
elif is_test:
    LOG_FILE = Path("/tmp/whis.test.log")
    config_toml = {}
else:
    XDG_CONFIG_HOME = Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")).expanduser()
    XDG_STATE_HOME = Path(os.environ.get("XDG_STATE_HOME", Path.home() / ".local" / "state")).expanduser()

    CONFIG_FILE = XDG_CONFIG_HOME / "whis" / "config.toml"

    LOG_DIR = XDG_STATE_HOME / "whis"
    LOG_DIR.mkdir(parents=True, exist_ok=True)
    LOG_FILE = str(LOG_DIR / "whis.log")

    # ~/.config/whis/config.toml might not exist yet, so we create it here with default values
    # from config.template.toml
    if not CONFIG_FILE.exists():
        template_text = (PACKAGE_ROOT / "config.template.toml").read_text()
        CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
        CONFIG_FILE.write_text(template_text)
    config_toml = tomllib.loads(CONFIG_FILE.read_text())


def get_cfg_var(key: str, default=None):
    """
    First try env var named as WHIS_<UPPER_SNAKE_KEY> then config.toml.
    """
    val = os.environ.get(f"WHIS_{key.upper()}")
    if val is not None:
        return val
    return config_toml.get(key, default)


llm_provider = get_cfg_var("llm_provider")  # ollama, openai...
llm_model = get_cfg_var("llm_model")  # gpt-3.5-turbo, gpt-4, qwen2:7b...
llm_temp = get_cfg_var("llm_temp")  # 0.2

LOG_FORMAT = "%(asctime)s - [%(levelname)s] - %(name)s: %(message)s"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
LOG_MAX_BYTES = 1024 * 1024 * 5  # 5 MB
LOG_BACKUP_COUNT = 5


def setup_logging():
    LOGGING_CONFIG = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {"format": LOG_FORMAT, "datefmt": LOG_DATE_FORMAT},
        },
        "handlers": {
            "console": {
                "level": "ERROR",  # it's a cli tool - no logs to the console
                "formatter": "standard",
                "class": "logging.StreamHandler",
            },
            "file": {
                "level": "DEBUG",
                "formatter": "standard",
                "class": "logging.handlers.RotatingFileHandler",
                "filename": LOG_FILE,
                "maxBytes": LOG_MAX_BYTES,
                "backupCount": LOG_BACKUP_COUNT,
                "encoding": "utf-8",  # todo is it necessary?
            },
        },
        "root": {
            "handlers": ["file", "console"],
            "level": "DEBUG",
        },
        "loggers": {
            # silence third-party loggers
            "urllib3": {"level": "WARNING", "propagate": True},
            "httpx": {"level": "WARNING", "propagate": True},
            "openai": {"level": "WARNING", "propagate": True},
            "ollama": {"level": "WARNING", "propagate": True},
            "httpcore": {"level": "WARNING", "propagate": True},
        },
    }
    logging.config.dictConfig(LOGGING_CONFIG)
    logging.captureWarnings(True)


def load_system_prompt():
    prompt = SYSTEM_PROMPT_FILE.read_text()
    context = {
        "current_datetime": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    }
    return prompt.format(**context)


def get_version() -> str:
    """Returns version. For development, we parse it from the pyproject.toml."""
    try:
        repo_root = Path(__file__).resolve().parents[1]
        pyproject = repo_root / "pyproject.toml"
        if pyproject.is_file():
            data = tomllib.loads(pyproject.read_text(encoding="utf-8"))
            return data.get("project", {}).get("version", "COULD_NOT_DETECT_VERSION")
    except Exception:
        pass

    try:
        return metadata.version("whis")
    except metadata.PackageNotFoundError:
        pass

    return "COULD_NOT_DETECT_VERSION"


class NotReadyError(Exception):
    """Raised when required config vars are not set"""

    missing_vars = []

    def __init__(self, message, missing_vars):
        super().__init__(message)
        self.missing_vars = missing_vars


REQUIRED_CONFIG_KEYS = ["llm_provider", "llm_model"]


def check_is_ready():
    missing_vars = []
    for var in REQUIRED_CONFIG_KEYS:
        if not get_cfg_var(var):
            missing_vars.append(var)
    if missing_vars:
        raise NotReadyError(f"Required config vars not set: {missing_vars}", missing_vars)


SYSTEM_PROMPT = load_system_prompt()



import logging
import subprocess
import time
import os
import sys
from . import config
import pyperclip

logger = logging.getLogger(__name__)


def paste_to_bash(text):
    """
    Forks a child process that pastes the given suggestion to the clipboard.
    Needs xclip and xdotool installed.
    """
    try:
        pid = os.fork()
        if pid > 0:
            return
    except OSError as e:
        print(f"Error: Fork failed: {e}", file=sys.stderr)
        return

    # give the shell some time
    time.sleep(0.1)
    pyperclip.copy(text)
    # simulate ctrl+shift+v (linux terminal pasting)
    subprocess.run(
        ["xdotool", "key", "--clearmodifiers", "ctrl+shift+v"],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


class ANSIColors:
    BOLD_CYAN = "\033[1;36m"
    BOLD_ORANGE = "\033[1;33m"
    GRAY = "\033[90m"
    DIM = "\033[2m"  # less visible than GRAY
    RESET = "\033[0m"


def colorize(text, color):
    return f"{color}{text}{ANSIColors.RESET}"


def get_whis_env_vars():
    return {k: v for k, v in os.environ.items() if k.startswith("WHIS_")}


def print_config():
    print(colorize(f"Python version: {sys.version}", ANSIColors.GRAY))
    print(colorize(f"ENV: {config.ENV}", ANSIColors.GRAY))
    print(colorize(f"version: {config.get_version()}", ANSIColors.GRAY))
    print(colorize(f"REPO_ROOT: {config.REPO_ROOT}", ANSIColors.GRAY))
    print(colorize(f"PACKAGE_ROOT: {config.PACKAGE_ROOT}", ANSIColors.GRAY))
    print(colorize(f"CONFIG_FILE: {config.CONFIG_FILE}", ANSIColors.GRAY))
    print(colorize(f"LOG_FILE: {config.LOG_FILE}", ANSIColors.GRAY))
    print(colorize(f"llm_provider: {config.llm_provider}", ANSIColors.GRAY))
    print(colorize(f"llm_model: {config.llm_model}", ANSIColors.GRAY))
    print(colorize(f"config_file_content: {config.config_toml}", ANSIColors.GRAY))
    env_vars_str = " ".join(f"{k}={v}" for k, v in get_whis_env_vars().items())
    print(colorize(f"ENV: {env_vars_str}", ANSIColors.GRAY))



import logging.config
from enum import Enum

from . import config
from .providers import registry
from .utils import paste_to_bash, ANSIColors, colorize

config.setup_logging()

logger = logging.getLogger(__name__)


class UserAction(Enum):
    EXECUTE = "_execute"
    QUIT = "_quit"
    RETRY = "_retry"
    FEEDBACK = "feedback"


class Session:
    def __init__(self):
        logger.info("WHIS %s", config.get_version())

        self.provider = registry.create_by_env()
        logger.info(self.provider)

    def oneshot(self, message: str):
        return paste_to_bash(self.provider.say(message))

    def run(self):
        logger.info("run")
        print(colorize(f"Whisperer: {self.provider}", ANSIColors.DIM))

        message = input("> ")

        while True:
            suggestion = self.provider.say(message)
            print(f"{colorize('? ' + suggestion, ANSIColors.BOLD_CYAN)}", end=" ")

            action = self._get_user_action()

            if action == UserAction.EXECUTE:
                logger.info("Pasting to bash: %s", suggestion)
                # pyperclip.copy(suggestion)
                paste_to_bash(suggestion)
                return

            elif action == UserAction.QUIT:
                print("Cancelled.")
                return
            elif action == UserAction.RETRY:
                message = "Try again, user wants something different."
            elif isinstance(action, str):  # custom feedback like "ok, but display human-readable file sizes"
                message = action

    def _get_user_action(self):
        print(colorize("[enter], [r]etry, feedback", f"{ANSIColors.DIM}{ANSIColors.GRAY}"))
        raw_input = input("> ").strip()
        logger.debug("User input: %s", raw_input)
        return self._parse_user_action(raw_input)

    def _parse_user_action(self, user_response):
        response_lower = user_response.lower()

        if response_lower in ("", "y", "yes", "ok"):
            _return = UserAction.EXECUTE
        elif response_lower in ("q", "quit"):
            _return = UserAction.QUIT
        elif response_lower in ("r", "retry"):
            _return = UserAction.RETRY
        else:
            _return = user_response  # custom feedback

        logger.debug("Parsed user action: %s", _return)
        return _return


def run_cli(oneshot=None):
    config.check_is_ready()
    session = Session()
    if not oneshot:
        session.run()
    else:
        session.oneshot(oneshot)
